Class: Innodb::Xdes

Inherits:
Object
  • Object
show all
Defined in:
lib/innodb/xdes.rb

Overview

An InnoDB "extent descriptor entry" or "+XDES+". These structures are used in the XDES entry array contained in FSP_HDR and XDES pages.

Note the distinction between XDES entries and XDES pages.

Constant Summary collapse

BITS_PER_PAGE =

Number of bits per page in the XDES entry bitmap field. Currently XDES entries store two bits per page, with the following meanings:

  • 1 = free (the page is free, or not in use)
  • 2 = clean (currently unused, always 1 when initialized)
2
BITMAP_BV_FREE =

The bit value for a free page.

1
BITMAP_BV_CLEAN =

The bit value for a clean page (currently unused in InnoDB).

2
BITMAP_BV_ALL =

The bitwise-OR of all bitmap bit values.

(BITMAP_BV_FREE | BITMAP_BV_CLEAN)
STATES =

The values used in the :state field indicating what the extent is used for (or what list it is on).

{
  1 => :free,       # The extent is completely empty and unused, and should
                    # be present on the filespace's FREE list.

  2 => :free_frag,  # Some pages of the extent are used individually, and
                    # the extent should be present on the filespace's
                    # FREE_FRAG list.

  3 => :full_frag,  # All pages of the extent are used individually, and
                    # the extent should be present on the filespace's
                    # FULL_FRAG list.

  4 => :fseg,       # The extent is wholly allocated to a file segment.
                    # Additional information about the state of this extent
                    # can be derived from the its presence on particular
                    # file segment lists (FULL, NOT_FULL, or FREE).
}

Instance Method Summary collapse

Constructor Details

#initialize(page, cursor) ⇒ Xdes

Returns a new instance of Xdes.



43
44
45
46
# File 'lib/innodb/xdes.rb', line 43

def initialize(page, cursor)
  @page = page
  @xdes = read_xdes_entry(page, cursor)
end

Instance Method Details

#==(other) ⇒ Object

Compare one Innodb::Xdes to another.



159
160
161
162
# File 'lib/innodb/xdes.rb', line 159

def ==(other)
  xdes[:this][:page] == other.xdes[:this][:page] &&
    xdes[:this][:offset] == other.xdes[:this][:offset]
end

#allocated_to_fseg?Boolean

Return whether this XDES entry is allocated to an fseg (the whole extent then belongs to the fseg).

Returns:

  • (Boolean)


90
91
92
# File 'lib/innodb/xdes.rb', line 90

def allocated_to_fseg?
  fseg_id != 0
end

#bitmapObject



86
# File 'lib/innodb/xdes.rb', line 86

def bitmap;     @xdes[:bitmap];     end

#each_page_statusObject

Iterate through all pages represented by this extent descriptor, yielding a page status hash for each page, containing the following fields:

:page   The page number.
:free   Boolean indicating whether the page is free.
:clean  Boolean indicating whether the page is clean (currently
      this bit is unused by InnoDB, and always set true).


109
110
111
112
113
114
115
116
117
118
119
120
121
122
123
124
125
126
127
128
129
# File 'lib/innodb/xdes.rb', line 109

def each_page_status
  unless block_given?
    return enum_for(:each_page_status)
  end

  bitmap = xdes[:bitmap].enum_for(:each_byte)

  bitmap.each_with_index do |byte, byte_index|
    (0..3).each_with_index do |page, page_index|
      page_number = xdes[:start_page] + (byte_index * 4) + page_index
      page_bits = ((byte >> (page * BITS_PER_PAGE)) & BITMAP_BV_ALL)
      page_status = {
        :free   => (page_bits & BITMAP_BV_FREE  != 0),
        :clean  => (page_bits & BITMAP_BV_CLEAN != 0),
      }
      yield page_number, page_status
    end
  end

  nil
end

#end_pageObject



81
# File 'lib/innodb/xdes.rb', line 81

def end_page;   @xdes[:end_page];   end

#free_pagesObject

Return the count of free pages (free bit is true) on this extent.



132
133
134
135
136
137
# File 'lib/innodb/xdes.rb', line 132

def free_pages
  each_page_status.inject(0) do |sum, (page_number, page_status)|
    sum += 1 if page_status[:free]
    sum
  end
end

#fseg_idObject



82
# File 'lib/innodb/xdes.rb', line 82

def fseg_id;    @xdes[:fseg_id];    end

#listObject



84
# File 'lib/innodb/xdes.rb', line 84

def list;       @xdes[:list];       end

#next_addressObject

Return the address of the next list pointer from the list node contained within the XDES entry. This is used by Innodb::List::Xdes to iterate through XDES entries in a list.



154
155
156
# File 'lib/innodb/xdes.rb', line 154

def next_address
  xdes[:list][:next]
end

#page_status(page_number) ⇒ Object

Return the status for a given page. This is relatively inefficient as implemented and could be done better.



96
97
98
99
# File 'lib/innodb/xdes.rb', line 96

def page_status(page_number)
  page_status_array = each_page_status.to_a
  page_status_array[page_number - xdes[:start_page]][1]
end

#prev_addressObject

Return the address of the previous list pointer from the list node contained within the XDES entry. This is used by Innodb::List::Xdes to iterate through XDES entries in a list.



147
148
149
# File 'lib/innodb/xdes.rb', line 147

def prev_address
  xdes[:list][:prev]
end

#read_xdes_entry(page, cursor) ⇒ Object

Read an XDES entry from a cursor.



59
60
61
62
63
64
65
66
67
68
69
70
71
72
73
# File 'lib/innodb/xdes.rb', line 59

def read_xdes_entry(page, cursor)
  extent_number = (cursor.position - page.pos_xdes_array) / size_entry
  start_page = page.offset + (extent_number * page.space.pages_per_extent)
  cursor.name("xdes[#{extent_number}]") do |c|
    {
      :start_page => start_page,
      :end_page   => start_page + page.space.pages_per_extent - 1,
      :fseg_id    => c.name("fseg_id") { c.get_uint64 },
      :this       => {:page => page.offset, :offset => c.position},
      :list       => c.name("list") { Innodb::List.get_node(c) },
      :state      => c.name("state") { STATES[c.get_uint32] },
      :bitmap     => c.name("bitmap") { c.get_bytes(size_bitmap) },
    }
  end
end

#size_bitmapObject

Size (in bytes) of the bitmap field in the XDES entry.



49
50
51
# File 'lib/innodb/xdes.rb', line 49

def size_bitmap
  (@page.space.pages_per_extent * BITS_PER_PAGE) / 8
end

#size_entryObject

Size (in bytes) of the an XDES entry.



54
55
56
# File 'lib/innodb/xdes.rb', line 54

def size_entry
  8 + Innodb::List::NODE_SIZE + 4 + size_bitmap
end

#start_pageObject



80
# File 'lib/innodb/xdes.rb', line 80

def start_page; @xdes[:start_page]; end

#stateObject



85
# File 'lib/innodb/xdes.rb', line 85

def state;      @xdes[:state];      end

#thisObject



83
# File 'lib/innodb/xdes.rb', line 83

def this;       @xdes[:this];       end

#used_pagesObject

Return the count of used pages (free bit is false) on this extent.



140
141
142
# File 'lib/innodb/xdes.rb', line 140

def used_pages
  @page.space.pages_per_extent - free_pages
end

#xdesObject

Return the stored extent descriptor entry.



76
77
78
# File 'lib/innodb/xdes.rb', line 76

def xdes
  @xdes
end